from copy import deepcopy
from unittest import IsolatedAsyncioTestCase

from marshmallow import ValidationError
from pyld import jsonld

from .......messaging.decorators.attach_decorator import AttachDecorator
from .......messaging.responder import BaseResponder, MockResponder
from .......resolver.did_resolver import DIDResolver
from .......storage.vc_holder.base import VCHolder
from .......storage.vc_holder.vc_record import VCRecord
from .......tests import mock
from .......utils.testing import create_test_profile
from .......vc.ld_proofs import DocumentLoader
from .......vc.vc_di.manager import VcDiManager
from .......vc.vc_ld.manager import VcLdpManager
from .......vc.vc_ld.validation_result import PresentationVerificationResult
from .....dif.pres_exch import SchemaInputDescriptor
from .....dif.pres_exch_handler import DIFPresExchError, DIFPresExchHandler
from .....dif.tests.test_data import (
    EXPANDED_CRED_FHIR_TYPE_1,
    EXPANDED_CRED_FHIR_TYPE_2,
    TEST_CRED_DICT,
)
from ....message_types import (
    ATTACHMENT_FORMAT,
    PRES_20,
    PRES_20_PROPOSAL,
    PRES_20_REQUEST,
)
from ....messages.pres import V20Pres
from ....messages.pres_format import V20PresFormat
from ....messages.pres_proposal import V20PresProposal
from ....messages.pres_request import V20PresRequest
from ....models.pres_exchange import V20PresExRecord
from ...handler import V20PresFormatHandlerError
from .. import handler as test_module
from ..handler import DIFPresFormatHandler

TEST_DID_SOV = "did:sov:LjgpST2rjsoxYegQDRm7EL"
TEST_DID_KEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"

DIF_PRES_REQUEST_A = {
    "presentation_definition": {
        "id": "32f54163-7166-48f1-93d8-ff217bdb0653",
        "submission_requirements": [
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "count": 2,
                "from": "A",
            }
        ],
        "input_descriptors": [
            {
                "id": "citizenship_input_1",
                "name": "EU Driver's License",
                "group": ["A"],
                "schema": [{"uri": "https://example.org/examples/degree.json"}],
                "constraints": {
                    "fields": [
                        {
                            "path": ["$.issuer.id", "$.vc.issuer.id", "$.issuer"],
                            "purpose": "The claim must be from one of the specified issuers",
                            "filter": {
                                "type": "string",
                                "enum": [
                                    "did:example:489398593",
                                    "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
                                    "did:sov:2wJPyULfLLnYTEFYzByfUR",
                                ],
                            },
                        }
                    ]
                },
            }
        ],
    }
}

DIF_PRES_REQUEST_B = {
    "options": {
        "challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7",
        "domain": "4jt78h47fh47",
    },
    "presentation_definition": {
        "id": "32f54163-7166-48f1-93d8-ff217bdb0654",
        "submission_requirements": [
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "min": 1,
                "from": "A",
            }
        ],
        "input_descriptors": [
            {
                "id": "citizenship_input_1",
                "name": "EU Driver's License",
                "group": ["A"],
                "schema": [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
                "constraints": {
                    "limit_disclosure": "required",
                    "fields": [
                        {
                            "path": ["$.credentialSubject.givenName"],
                            "purpose": "The claim must be from one of the specified issuers",
                            "filter": {
                                "type": "string",
                                "enum": ["JOHN", "CAI"],
                            },
                        }
                    ],
                },
            }
        ],
    },
}

DIF_PRES_REQUEST_SEQUENCE = {
    "options": {
        "challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7",
        "domain": "4jt78h47fh47",
    },
    "presentation_definition": {
        "id": "32f54163-7166-48f1-93d8-ff217bdb0654",
        "submission_requirements": [
            {
                "name": "Citizenship Information",
                "rule": "all",
                "from": "A",
            },
            {
                "name": "Citizenship Information v2",
                "rule": "pick",
                "min": 1,
                "from": "B",
            },
        ],
        "input_descriptors": [
            {
                "id": "citizenship_input_1",
                "name": "EU Driver's License",
                "group": ["A"],
                "schema": [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
                "constraints": {
                    "limit_disclosure": "required",
                    "fields": [
                        {
                            "path": ["$.credentialSubject.givenName"],
                            "purpose": "The claim must be from one of the specified issuers",
                            "filter": {
                                "type": "string",
                                "enum": ["JOHN", "CAI"],
                            },
                        }
                    ],
                },
            },
            {
                "id": "citizenship_input_2",
                "name": "EU Driver's License v2",
                "group": ["B"],
                "schema": [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
                "constraints": {
                    "limit_disclosure": "required",
                    "fields": [
                        {
                            "path": ["$.credentialSubject.givenName"],
                            "purpose": "The claim must be from one of the specified issuers",
                            "filter": {
                                "type": "string",
                                "enum": ["JOHN", "CAI"],
                            },
                        }
                    ],
                },
            },
        ],
    },
}

DIF_PRES_PROPOSAL = {
    "input_descriptors": [
        {
            "id": "citizenship_input_1",
            "name": "EU Driver's License",
            "group": ["A"],
            "schema": [
                {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
            ],
            "constraints": {
                "fields": [
                    {
                        "path": ["$.issuer.id", "$.vc.issuer.id", "$.issuer"],
                        "purpose": "The claim must be from one of the specified issuers",
                        "filter": {
                            "type": "string",
                            "enum": [
                                "did:example:489398593",
                                "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
                                "did:sov:2wJPyULfLLnYTEFYzByfUR",
                            ],
                        },
                    }
                ]
            },
        }
    ]
}

DIF_PRES = {
    "@context": ["https://www.w3.org/2018/credentials/v1"],
    "type": ["VerifiablePresentation"],
    "verifiableCredential": [
        {
            "@context": [
                "https://www.w3.org/2018/credentials/v1",
                "https://w3id.org/citizenship/v1",
                "https://w3id.org/security/bbs/v1",
            ],
            "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
            "type": ["PermanentResidentCard", "VerifiableCredential"],
            "credentialSubject": {
                "id": "did:example:b34ca6cd37bbf23",
                "type": ["Person", "PermanentResident"],
                "givenName": "JOHN",
            },
            "issuanceDate": "2010-01-01T19:53:24Z",
            "issuer": "did:key:zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v",
            "proof": {
                "type": "BbsBlsSignatureProof2020",
                "nonce": "3AuruhJQrXtEgiagiJ+FwVf2S0SnzUDJvnO61YecQsJ7ImR1mPcoVjJJ0HOhfkFpoYI=",
                "proofValue": "ABkBuAaPlP5A7JWY78Xf69oBnsMLcD1RXbIFYhcLoXPXW12CG9glnnqnPLsGri5xsA3LcP0kg74X+sAjKXGRGy3uvp412Dm0FuohYNboQcLne5KOAa5AxU4bjmwQsxdfduVqhriro1N+YTkuB4SMmO/5ooL0N3OHsYdExg7nSzWqmZoqgp+3CwIxF0a/oyKTcxJORuIqAAAAdInlL9teSIX49NJGEZfBO7IrdjT2iggH/G0AlPWoEvrWIbuCRQ69K83n5o7oJVjqhAAAAAIaVmlAD6+FEKA4eg0OaWOKPrd5Kq8rv0vIwjJ71egxll0Fqq4zDWQ/+yl3Pteh0Wyuyvpm19/sj6tiCWj4PkA+rpxtR2bXpnrCTKUffFFNBjVvVziXDS0KWkGUB7XU9mjUa4USC7Iub3bZZCnFjQA5AAAADzkGwGD837r33e7OTrGEti8eAkvFDcyCgA4ck/X+5HJjAJclHWbl4SNQR8CiNZyzJpvxW+jbNBcwmEvocYArddk3F78Ki0Qnp6aU9eDgfOOx1iW2BXLUjrhq5I2hP5/WQF3CEDYRjczGjzM9T8/coeC36YAp0zJunIXUKb8SPDSOISafibYRYFB4xhlWKXWloDelafyujOBST8KZNM8FmF4DSbXrO8vmZbjuR/8ntUcUK7X2rNbuZ3M5eWZDF8pL+SA9gQitKfPHEocoYAdhgEAM7ZNAJ+TgOcx9gtZIhDWKDNnFxIeoOAylbD1xZd9xbWtq3Bk3R79xqsKxFRJRNxk/9b6fJruP292+qM5lxcZ1jUz/dJUYFI93hH4Mso75CjGRN78MAY9SNifl6H8qcxTpBn4332LlFhRznLbtnc4YSWA/fvVqaN9h2zCH/6AdbLKXGffV34EF7DadwJsi9jsc+YlSMn6qaIUIDTdGLwh4KKpSH5bVbg/mVCcXPTJplFgYwRsOdiQbZY/740dJyo1lPjQ0Lvdio8W2M8c73ujeJU70CNLkgjJAMUPGrCFtGxBH2eeLBQ0P95qRZAIcJ7U0MibZLaRjoUOuTla5BIt2038PJ6XhcY6BEJaLyJOPEQ==",
                "verificationMethod": "did:key:zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v#zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v",
                "proofPurpose": "assertionMethod",
                "created": "2021-05-05T15:22:30.523465",
            },
        }
    ],
    "presentation_submission": {
        "id": "a5fcfe44-2c30-497d-af02-98e539da9a0f",
        "definition_id": "32f54163-7166-48f1-93d8-ff217bdb0653",
        "descriptor_map": [
            {
                "id": "citizenship_input_1",
                "format": "ldp_vc",
                "path": "$.verifiableCredential[0]",
            }
        ],
    },
    "proof": {
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:sov:4QxzWk3ajdnEA37NdNU5Kt#key-1",
        "created": "2021-05-05T15:23:03.023971",
        "proofPurpose": "authentication",
        "challenge": "40429d49-5e8f-4ffc-baf8-e332412f1247",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..2uBYmg7muE9ZPVeAGo_ibVfLkCjf2hGshr2o5i8pAwFyNBM-kDHXofuq1MzJgb19wzb01VIu91hY_ajjt9KFAA",
    },
}

DIF_PRES_SEQUENCE = [
    DIF_PRES,
    {
        "@context": ["https://www.w3.org/2018/credentials/v1"],
        "type": ["VerifiablePresentation"],
        "verifiableCredential": [
            {
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://w3id.org/citizenship/v1",
                    "https://w3id.org/security/bbs/v1",
                ],
                "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
                "type": ["PermanentResidentCard", "VerifiableCredential"],
                "credentialSubject": {
                    "id": "did:example:b34ca6cd37bbf23",
                    "type": ["Person", "PermanentResident"],
                    "givenName": "JOHN",
                },
                "issuanceDate": "2010-01-01T19:53:24Z",
                "issuer": "did:key:zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v",
                "proof": {
                    "type": "BbsBlsSignatureProof2020",
                    "nonce": "3AuruhJQrXtEgiagiJ+FwVf2S0SnzUDJvnO61YecQsJ7ImR1mPcoVjJJ0HOhfkFpoYI=",
                    "proofValue": "ABkBuAaPlP5A7JWY78Xf69oBnsMLcD1RXbIFYhcLoXPXW12CG9glnnqnPLsGri5xsA3LcP0kg74X+sAjKXGRGy3uvp412Dm0FuohYNboQcLne5KOAa5AxU4bjmwQsxdfduVqhriro1N+YTkuB4SMmO/5ooL0N3OHsYdExg7nSzWqmZoqgp+3CwIxF0a/oyKTcxJORuIqAAAAdInlL9teSIX49NJGEZfBO7IrdjT2iggH/G0AlPWoEvrWIbuCRQ69K83n5o7oJVjqhAAAAAIaVmlAD6+FEKA4eg0OaWOKPrd5Kq8rv0vIwjJ71egxll0Fqq4zDWQ/+yl3Pteh0Wyuyvpm19/sj6tiCWj4PkA+rpxtR2bXpnrCTKUffFFNBjVvVziXDS0KWkGUB7XU9mjUa4USC7Iub3bZZCnFjQA5AAAADzkGwGD837r33e7OTrGEti8eAkvFDcyCgA4ck/X+5HJjAJclHWbl4SNQR8CiNZyzJpvxW+jbNBcwmEvocYArddk3F78Ki0Qnp6aU9eDgfOOx1iW2BXLUjrhq5I2hP5/WQF3CEDYRjczGjzM9T8/coeC36YAp0zJunIXUKb8SPDSOISafibYRYFB4xhlWKXWloDelafyujOBST8KZNM8FmF4DSbXrO8vmZbjuR/8ntUcUK7X2rNbuZ3M5eWZDF8pL+SA9gQitKfPHEocoYAdhgEAM7ZNAJ+TgOcx9gtZIhDWKDNnFxIeoOAylbD1xZd9xbWtq3Bk3R79xqsKxFRJRNxk/9b6fJruP292+qM5lxcZ1jUz/dJUYFI93hH4Mso75CjGRN78MAY9SNifl6H8qcxTpBn4332LlFhRznLbtnc4YSWA/fvVqaN9h2zCH/6AdbLKXGffV34EF7DadwJsi9jsc+YlSMn6qaIUIDTdGLwh4KKpSH5bVbg/mVCcXPTJplFgYwRsOdiQbZY/740dJyo1lPjQ0Lvdio8W2M8c73ujeJU70CNLkgjJAMUPGrCFtGxBH2eeLBQ0P95qRZAIcJ7U0MibZLaRjoUOuTla5BIt2038PJ6XhcY6BEJaLyJOPEQ==",
                    "verificationMethod": "did:key:zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v#zUC74bgefTdc43KS1psXgXf4jLaHyaj2qCQqQTXrtmSYGf1PxiJhrH6LGpaBMyj6tqAKmjGyMaS4RfNo2an77vT1HfzJUNPk4H7TCuJvSp4vet4Cu67kn2JSegoQNFSA1tbwU8v",
                    "proofPurpose": "assertionMethod",
                    "created": "2021-05-05T15:22:30.523465",
                },
            }
        ],
        "presentation_submission": {
            "id": "a5fcfe44-2c30-497d-af02-98e539da9a0f",
            "definition_id": "32f54163-7166-48f1-93d8-ff217bdb0653",
            "descriptor_map": [
                {
                    "id": "citizenship_input_2",
                    "format": "ldp_vp",
                    "path": "$.verifiableCredential[0]",
                }
            ],
        },
        "proof": {
            "type": "Ed25519Signature2018",
            "verificationMethod": "did:sov:4QxzWk3ajdnEA37NdNU5Kt#key-1",
            "created": "2021-05-05T15:23:03.023971",
            "proofPurpose": "authentication",
            "challenge": "40429d49-5e8f-4ffc-baf8-e332412f1247",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..2uBYmg7muE9ZPVeAGo_ibVfLkCjf2hGshr2o5i8pAwFyNBM-kDHXofuq1MzJgb19wzb01VIu91hY_ajjt9KFAA",
        },
    },
]


TEST_CRED = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/citizenship/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
    "type": ["VerifiableCredential", "PermanentResidentCard"],
    "issuer": "did:key:zUC76eX863NT1BEFYEfSMRY4CSVwRmRdKBtv3cwTwytDXNEAvJxFr3GBjkpKRs3xg9FznUhxXkwDLLu7UjDnetSQNvGiT1ivHdDFByZdXoWLYpDDRph5eZDXTGNweuq3Z8uAo3o",
    "identifier": "83627465",
    "name": "Permanent Resident Card",
    "description": "Government of Example Permanent Resident Card.",
    "issuanceDate": "2010-01-01T19:53:24Z",
    "expirationDate": "2029-12-03T12:19:52Z",
    "credentialSubject": {
        "id": "did:sov:4QxzWk3ajdnEA37NdNU5Kt",
        "type": ["PermanentResident", "Person"],
        "givenName": "JOHN",
        "familyName": "SMITH",
        "gender": "Male",
        "image": "data:image/png;base64,iVBORw0KGgokJggg==",
        "residentSince": "2015-01-01",
        "lprCategory": "C09",
        "lprNumber": "999-999-999",
        "commuterClassification": "C1",
        "birthCountry": "Bahamas",
        "birthDate": "1958-07-17",
    },
}


class TestDIFFormatHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

        # Set custom document loader
        self.profile.context.injector.bind_instance(DIDResolver, DIDResolver([]))
        self.profile.context.injector.bind_instance(
            DocumentLoader, DocumentLoader(self.profile)
        )
        self.profile.context.injector.bind_instance(BaseResponder, MockResponder())

        self.manager = VcLdpManager(self.profile)
        self.profile.context.injector.bind_instance(VcLdpManager, self.manager)

        self.handler = DIFPresFormatHandler(self.profile)
        assert self.handler.profile

    def test_validate_fields(self):
        self.handler.validate_fields(PRES_20, DIF_PRES)
        self.handler.validate_fields(PRES_20_PROPOSAL, DIF_PRES_PROPOSAL)
        self.handler.validate_fields(PRES_20_REQUEST, DIF_PRES_REQUEST_A)
        self.handler.validate_fields(PRES_20_REQUEST, DIF_PRES_REQUEST_B)

        with self.assertRaises(ValidationError):
            incorrect_pres = DIF_PRES.copy()
            incorrect_pres.pop("@context")
            self.handler.validate_fields(PRES_20, incorrect_pres)

    async def test_create_bound_request_a(self):
        dif_proposal_dict = {
            "input_descriptors": [
                {
                    "id": "citizenship_input_1",
                    "name": "EU Driver's License",
                    "group": ["A"],
                    "schema": [
                        {
                            "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                        }
                    ],
                    "constraints": {
                        "limit_disclosure": "required",
                        "fields": [
                            {
                                "path": ["$.credentialSubject.givenName"],
                                "purpose": "The claim must be from one of the specified issuers",
                                "filter": {"type": "string", "enum": ["JOHN", "CAI"]},
                            }
                        ],
                    },
                }
            ]
        }
        dif_pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            proposals_attach=[AttachDecorator.data_json(dif_proposal_dict, ident="dif")],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_proposal=dif_pres_proposal,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        output = await self.handler.create_bound_request(pres_ex_record=record)
        assert isinstance(output[0], V20PresFormat) and isinstance(
            output[1], AttachDecorator
        )

    async def test_create_bound_request_b(self):
        dif_proposal_dict = {
            "options": {"challenge": "test123"},
            "input_descriptors": [
                {
                    "id": "citizenship_input_1",
                    "name": "EU Driver's License",
                    "group": ["A"],
                    "schema": [
                        {
                            "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                        }
                    ],
                    "constraints": {
                        "limit_disclosure": "required",
                        "fields": [
                            {
                                "path": ["$.credentialSubject.givenName"],
                                "purpose": "The claim must be from one of the specified issuers",
                                "filter": {"type": "string", "enum": ["JOHN", "CAI"]},
                            }
                        ],
                    },
                }
            ],
        }
        dif_pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            proposals_attach=[AttachDecorator.data_json(dif_proposal_dict, ident="dif")],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_proposal=dif_pres_proposal,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        output = await self.handler.create_bound_request(pres_ex_record=record)
        assert isinstance(output[0], V20PresFormat) and isinstance(
            output[1], AttachDecorator
        )

    async def test_create_bound_request_c(self):
        dif_proposal_dict = {
            "options": {"domain": "test123"},
            "input_descriptors": [
                {
                    "id": "citizenship_input_1",
                    "name": "EU Driver's License",
                    "group": ["A"],
                    "schema": [
                        {
                            "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                        }
                    ],
                    "constraints": {
                        "limit_disclosure": "required",
                        "fields": [
                            {
                                "path": ["$.credentialSubject.givenName"],
                                "purpose": "The claim must be from one of the specified issuers",
                                "filter": {"type": "string", "enum": ["JOHN", "CAI"]},
                            }
                        ],
                    },
                }
            ],
        }
        dif_pres_proposal = V20PresProposal(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            proposals_attach=[AttachDecorator.data_json(dif_proposal_dict, ident="dif")],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_proposal=dif_pres_proposal,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        output = await self.handler.create_bound_request(pres_ex_record=record)
        assert isinstance(output[0], V20PresFormat) and isinstance(
            output[1], AttachDecorator
        )

    async def test_create_pres(self):
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, {})
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_pd_schema_uri(self):
        dif_pres_req = deepcopy(DIF_PRES_REQUEST_B)
        dif_pres_req["presentation_definition"]["input_descriptors"][0]["schema"][0][
            "uri"
        ] = "test.json"
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(dif_pres_req, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        request_data = {}
        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, request_data)
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_prover_proof_spec(self):
        dif_pres_spec = deepcopy(DIF_PRES_REQUEST_A)
        dif_pres_spec["issuer_id"] = "test123"
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        request_data = {}
        request_data["dif"] = dif_pres_spec
        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, request_data)
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_prover_proof_spec_with_record_ids(self):
        dif_pres_spec = deepcopy(DIF_PRES_REQUEST_A)
        dif_pres_spec["issuer_id"] = "test123"
        dif_pres_spec["record_ids"] = {"test_input_descriptor_id": ["test1"]}
        cred_list = [
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="did:example:489398593",
                subject_ids=[
                    "did:sov:WgWxqztrNooG92RXvxSTWv",
                ],
                proof_types=["Ed25519Signature2018"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...", "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test1",
            )
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        request_data = {}
        request_data["dif"] = dif_pres_spec

        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(fetch=mock.CoroutineMock(return_value=cred_list))
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with (
            mock.patch.object(
                DIFPresExchHandler,
                "create_vp",
                mock.CoroutineMock(),
            ) as mock_create_vp,
            mock.patch.object(
                VCHolder,
                "search_credentials",
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        fetch=mock.CoroutineMock(return_value=cred_list)
                    )
                ),
            ),
        ):
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, request_data)
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_prover_proof_spec_with_reveal_doc(self):
        dif_pres_spec = deepcopy(DIF_PRES_REQUEST_A)
        dif_pres_spec["issuer_id"] = "test123"
        dif_pres_spec["reveal_doc"] = {
            "@context": [
                "https://www.w3.org/2018/credentials/v1",
                "https://w3id.org/security/bbs/v1",
            ],
            "type": ["VerifiableCredential", "LabReport"],
            "@explicit": True,
            "@requireAll": True,
            "issuanceDate": {},
            "issuer": {},
            "credentialSubject": {
                "Observation": [
                    {"effectiveDateTime": {}, "@explicit": True, "@requireAll": True}
                ],
                "@explicit": True,
                "@requireAll": True,
            },
        }
        cred_list = [
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="did:example:489398593",
                subject_ids=[
                    "did:sov:WgWxqztrNooG92RXvxSTWv",
                ],
                proof_types=["Ed25519Signature2018"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...", "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test1",
            )
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        request_data = {}
        request_data["dif"] = dif_pres_spec

        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(fetch=mock.CoroutineMock(return_value=cred_list))
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, request_data)
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_one_of_filter(self):
        cred_list = [
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="did:example:489398593",
                subject_ids=[
                    "did:sov:WgWxqztrNooG92RXvxSTWv",
                ],
                proof_types=["Ed25519Signature2018"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...", "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test1",
            )
        ]
        pres_request = deepcopy(DIF_PRES_REQUEST_B)
        pres_request["presentation_definition"]["input_descriptors"][0]["schema"] = {
            "oneof_filter": [
                [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
                [{"uri": "https://www.w3.org/Test#Test"}],
            ]
        }
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(pres_request, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        request_data = {"dif": {}}

        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(fetch=mock.CoroutineMock(return_value=cred_list))
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, request_data)
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_no_challenge(self):
        dif_pres_req = deepcopy(DIF_PRES_REQUEST_B)
        del dif_pres_req["options"]["challenge"]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(dif_pres_req, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        request_data = {}
        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, request_data)
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_storage_not_found(self):
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(side_effect=test_module.StorageNotFoundError())
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(V20PresFormatHandlerError):
            await self.handler.create_pres(record)

    async def test_create_pres_pd_claim_format_ed255(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["Ed25519Signature2018"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, {})
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_pd_claim_format_secp256r1(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["EcdsaSecp256r1Signature2019"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, {})
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_pd_claim_format_bls12381g2(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["BbsBlsSignature2020"]}
        }
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, {})
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_verify_pres_sequence(self):
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_SEQUENCE, ident="dif")
            ],
        )
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_SEQUENCE, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            self.manager,
            "verify_presentation",
            mock.CoroutineMock(
                return_value=PresentationVerificationResult(verified=True)
            ),
        ):
            output = await self.handler.verify_pres(record)
            assert output.verified

        with mock.patch.object(
            self.manager,
            "verify_presentation",
            mock.CoroutineMock(
                return_value=PresentationVerificationResult(verified=False)
            ),
        ):
            output = await self.handler.verify_pres(record)
            assert output.verified == "false"

    async def test_verify_pres(self):
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[AttachDecorator.data_json(DIF_PRES, ident="dif")],
        )
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            self.manager,
            "verify_presentation",
            mock.CoroutineMock(
                return_value=PresentationVerificationResult(verified=True)
            ),
        ):
            output = await self.handler.verify_pres(record)
            assert output.verified

    async def test_verify_pres_no_challenge(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        del test_pd["options"]["challenge"]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[AttachDecorator.data_json(DIF_PRES, ident="dif")],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        assert await self.handler.verify_pres(record)

    async def test_create_pres_cred_limit_disclosure_no_bbs(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["Ed25519Signature2018"]}
        }
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with self.assertRaises(V20PresFormatHandlerError):
            await self.handler.create_pres(record)

    async def test_create_pres_cred_claim_format_single_ldp_vp_error(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["test"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with self.assertRaises(V20PresFormatHandlerError):
            await self.handler.create_pres(record)

    async def test_create_pres_cred_claim_format_double_ldp_vp_error(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["test1", "test2"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with self.assertRaises(V20PresFormatHandlerError):
            await self.handler.create_pres(record)

    async def test_create_pres_cred_no_tag_query(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["input_descriptors"][0]["schema"][0][
            "required"
        ] = False
        test_pd["presentation_definition"]["input_descriptors"][0]["schema"][1][
            "required"
        ] = False
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with mock.patch.object(
            DIFPresExchHandler,
            "create_vp",
            mock.CoroutineMock(),
        ) as mock_create_vp:
            mock_create_vp.return_value = DIF_PRES
            output = await self.handler.create_pres(record, {})
            assert isinstance(output[0], V20PresFormat) and isinstance(
                output[1], AttachDecorator
            )
            assert output[1].data.json_ == DIF_PRES

    async def test_create_pres_cred_claim_format_no_ldp_vp(self):
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vc": {"proof_type": ["test"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )

        with self.assertRaises(V20PresFormatHandlerError):
            await self.handler.create_pres(record)

    async def test_process_vcrecords_return_list(self):
        cred_list = [
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="https://example.edu/issuers/565049",
                subject_ids=[
                    "did:sov:LjgpST2rjsoxYegQDRm7EL",
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                ],
                proof_types=["BbsBlsSignature2020"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...": "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test1",
            ),
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="https://example.edu/issuers/565049",
                subject_ids=[
                    "did:sov:LjgpST2rjsoxYegQDRm7EL",
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                ],
                proof_types=["BbsBlsSignature2020"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...": "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test2",
            ),
        ]
        record_ids = {"test1"}
        (
            returned_cred_list,
            returned_record_ids,
        ) = await self.handler.process_vcrecords_return_list(cred_list, record_ids)
        assert len(returned_cred_list) == 1
        assert len(returned_record_ids) == 2
        assert returned_cred_list[0].record_id == "test2"

    async def test_retrieve_uri_list_from_schema_filter(self):
        test_schema_filter = [
            [
                SchemaInputDescriptor(uri="test123"),
                SchemaInputDescriptor(uri="test321", required=True),
            ]
        ]
        test_one_of_uri_groups = await self.handler.retrieve_uri_list_from_schema_filter(
            test_schema_filter
        )
        assert test_one_of_uri_groups == [["test123", "test321"]]

    async def test_verify_received_pres_a(self):
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=DIF_PRES,
                    ident="dif",
                )
            ],
        )
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)

    async def test_verify_received_pres_b(self):
        dif_proof = deepcopy(DIF_PRES)
        dif_proof["verifiableCredential"][0]["credentialSubject"]["givenName"] = {
            "test": "JOHN"
        }
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )

        dif_proof_req = deepcopy(DIF_PRES_REQUEST_B)
        dif_proof_req["presentation_definition"]["input_descriptors"][0]["constraints"][
            "fields"
        ][0]["path"] = ["$.credentialSubject.givenName.test"]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(dif_proof_req, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)

    async def test_verify_received_pres_c(self):
        dif_proof = deepcopy(DIF_PRES)
        dif_proof["verifiableCredential"][0]["credentialSubject"]["givenName"] = {
            "test": "JOHN"
        }
        dif_proof["verifiableCredential"][0]["credentialSubject"]["test"] = {
            "test1": {"test2": "TEST"}
        }
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )

        dif_proof_req = deepcopy(DIF_PRES_REQUEST_B)
        dif_proof_req["presentation_definition"]["input_descriptors"][0][
            "constraints"
        ] = {
            "limit_disclosure": "required",
            "fields": [
                {
                    "path": ["$.credentialSubject.givenName.test"],
                    "filter": {
                        "type": "string",
                        "enum": ["JOHN", "CAI"],
                    },
                },
                {
                    "path": ["$.credentialSubject.test.test1.test2"],
                    "filter": {
                        "type": "string",
                        "enum": ["TEST"],
                    },
                },
            ],
        }
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(dif_proof_req, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)

    async def test_verify_received_pres_sequence(self):
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=DIF_PRES_SEQUENCE,
                    ident="dif",
                )
            ],
        )
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_SEQUENCE, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)

    async def test_verify_received_limit_disclosure_a(self):
        dif_proof = deepcopy(DIF_PRES)
        cred_dict = deepcopy(TEST_CRED_DICT)
        cred_dict["credentialSubject"]["Patient"] = [
            {
                "address": [
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                ]
            },
            {
                "address": [
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                ]
            },
        ]
        dif_proof["verifiableCredential"] = []
        dif_proof["verifiableCredential"].append(cred_dict)
        dif_proof["verifiableCredential"].append(cred_dict)
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )
        pres_request = deepcopy(DIF_PRES_REQUEST_B)
        pres_request["presentation_definition"]["input_descriptors"][0]["constraints"] = {
            "limit_disclosure": "required",
            "fields": [
                {
                    "path": ["$.credentialSubject.Patient[0].address[*].city"],
                    "purpose": "Test",
                }
            ],
        }
        pres_request["presentation_definition"]["input_descriptors"][0]["schema"] = {
            "oneof_filter": [
                [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://www.vdel.com/MedicalPass"},
                    {"uri": "http://hl7.org/fhir/Patient"},
                ],
                [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
            ]
        }
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(pres_request, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(jsonld, "expand", mock.MagicMock()) as mock_jsonld_expand:
            mock_jsonld_expand.return_value = EXPANDED_CRED_FHIR_TYPE_2
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)

    async def test_verify_received_limit_disclosure_b(self):
        dif_proof = deepcopy(DIF_PRES)
        cred_dict = deepcopy(TEST_CRED_DICT)
        cred_dict["credentialSubject"]["Patient"]["address"] = [
            {
                "@id": "urn:bnid:_:c14n1",
                "city": "Рума",
            },
            {
                "@id": "urn:bnid:_:c14n1",
                "city": "Рума",
            },
        ]
        dif_proof["verifiableCredential"] = []
        dif_proof["verifiableCredential"].append(cred_dict)
        dif_proof["verifiableCredential"].append(cred_dict)
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )
        pres_request = deepcopy(DIF_PRES_REQUEST_B)
        pres_request["presentation_definition"]["input_descriptors"][0]["constraints"] = {
            "limit_disclosure": "required",
            "fields": [
                {
                    "path": ["$.credentialSubject.Patient[*].address"],
                    "purpose": "Test",
                }
            ],
        }
        pres_request["presentation_definition"]["input_descriptors"][0]["schema"] = [
            {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
            {"uri": "https://www.vdel.com/MedicalPass"},
            {"uri": "http://hl7.org/fhir/Patient"},
        ]
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(pres_request, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(jsonld, "expand", mock.MagicMock()) as mock_jsonld_expand:
            mock_jsonld_expand.return_value = EXPANDED_CRED_FHIR_TYPE_1
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)

    async def test_verify_received_pres_invalid_jsonpath(self):
        dif_proof = deepcopy(DIF_PRES)
        dif_proof["presentation_submission"]["descriptor_map"][0]["path"] = (
            "$.verifiableCredential[1]"
        )
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )

        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(
            test_module.LOGGER, "error", mock.MagicMock()
        ) as mock_log_err:
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)
            mock_log_err.assert_called_once()

    async def test_verify_received_pres_no_match_a(self):
        dif_proof_req = deepcopy(DIF_PRES_REQUEST_B)
        dif_proof_req["presentation_definition"]["input_descriptors"][0]["constraints"][
            "fields"
        ][0]["path"] = ["$.credentialSubject.givenName2"]

        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=DIF_PRES,
                    ident="dif",
                )
            ],
        )

        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(dif_proof_req, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(
            test_module.LOGGER, "error", mock.MagicMock()
        ) as mock_log_err:
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)
            mock_log_err.assert_called_once()

    async def test_verify_received_pres_no_match_b(self):
        dif_proof_req = deepcopy(DIF_PRES_REQUEST_B)
        dif_proof_req["presentation_definition"]["input_descriptors"][0]["constraints"][
            "fields"
        ][0]["filter"]["enum"] = ["Test"]

        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=DIF_PRES,
                    ident="dif",
                )
            ],
        )

        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(dif_proof_req, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(
            test_module.LOGGER, "error", mock.MagicMock()
        ) as mock_log_err:
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)
            mock_log_err.assert_called_once()

    async def test_verify_received_pres_limit_disclosure_fail_a(self):
        dif_proof = deepcopy(DIF_PRES)
        dif_proof["verifiableCredential"][0]["expirationDate"] = "Test"
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )

        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(
            test_module.LOGGER, "error", mock.MagicMock()
        ) as mock_log_err:
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)
            mock_log_err.assert_called_once()

    async def test_verify_received_pres_limit_disclosure_fail_b(self):
        dif_proof = deepcopy(DIF_PRES)
        dif_proof["verifiableCredential"][0]["credentialSubject"]["test"] = "Test"
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )

        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(DIF_PRES_REQUEST_B, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(
            test_module.LOGGER, "error", mock.MagicMock()
        ) as mock_log_err:
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)
            mock_log_err.assert_called_once()

    async def test_verify_received_pres_fail_schema_filter(self):
        dif_proof = deepcopy(DIF_PRES)
        cred_dict = deepcopy(TEST_CRED_DICT)
        cred_dict["credentialSubject"]["Patient"] = [
            {
                "address": [
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                ]
            },
            {
                "address": [
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                    {
                        "@id": "urn:bnid:_:c14n1",
                        "city": "Рума",
                    },
                ]
            },
        ]
        dif_proof["verifiableCredential"] = []
        dif_proof["verifiableCredential"].append(cred_dict)
        dif_proof["verifiableCredential"].append(cred_dict)
        dif_pres = V20Pres(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20][V20PresFormat.Format.DIF.api],
                )
            ],
            presentations_attach=[
                AttachDecorator.data_json(
                    mapping=dif_proof,
                    ident="dif",
                )
            ],
        )
        pres_request = deepcopy(DIF_PRES_REQUEST_B)
        pres_request["presentation_definition"]["input_descriptors"][0]["constraints"] = {
            "limit_disclosure": "required",
            "fields": [
                {
                    "path": ["$.credentialSubject.Patient[0].address[*].city"],
                    "purpose": "Test",
                }
            ],
        }
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(pres_request, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            pres=dif_pres,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with (
            mock.patch.object(
                test_module.LOGGER, "error", mock.MagicMock()
            ) as mock_log_err,
            mock.patch.object(jsonld, "expand", mock.MagicMock()) as mock_jsonld_expand,
        ):
            mock_jsonld_expand.return_value = EXPANDED_CRED_FHIR_TYPE_2
            await self.handler.receive_pres(message=dif_pres, pres_ex_record=record)
            mock_log_err.assert_called_once()

    async def test_create_pres_catch_typeerror(self):
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(side_effect=TypeError)
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        await self.handler.create_pres(record)

    async def test_create_pres_catch_diferror(self):
        cred_list = [
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="did:example:489398593",
                subject_ids=[
                    "did:sov:WgWxqztrNooG92RXvxSTWv",
                ],
                proof_types=["Ed25519Signature2018"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...", "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test1",
            )
        ]
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(fetch=mock.CoroutineMock(return_value=cred_list))
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)
        test_pd = deepcopy(DIF_PRES_REQUEST_B)
        dif_pres_request = V20PresRequest(
            formats=[
                V20PresFormat(
                    attach_id="dif",
                    format_=ATTACHMENT_FORMAT[PRES_20_REQUEST][
                        V20PresFormat.Format.DIF.api
                    ],
                )
            ],
            request_presentations_attach=[
                AttachDecorator.data_json(test_pd, ident="dif")
            ],
        )
        record = V20PresExRecord(
            pres_ex_id="pxid",
            thread_id="thid",
            connection_id="conn_id",
            initiator="init",
            role="role",
            state="state",
            pres_request=dif_pres_request,
            verified="false",
            auto_present=True,
            error_msg="error",
        )
        with mock.patch.object(
            DIFPresExchHandler, "create_vp", mock.MagicMock()
        ) as mock_create_vp:
            mock_create_vp.side_effect = DIFPresExchError("TEST")
            await self.handler.create_pres(record)

    async def test_get_type_manager_options(self):
        self.profile = await create_test_profile()
        handler = DIFPresFormatHandler(self.profile)
        dif_proof = {"proof": {"type": "DataIntegrityProof"}}
        pres_request = {"options": {"challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7"}}
        manager, options = handler._get_type_manager_options(dif_proof, pres_request)
        assert isinstance(manager, VcDiManager)
        assert options == pres_request

        dif_proof = {"proof": {"type": "LDPProof"}}
        pres_request = {"options": {"challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7"}}
        manager, options = handler._get_type_manager_options(dif_proof, pres_request)
        assert isinstance(manager, VcLdpManager)
        assert options.challenge == "3fa85f64-5717-4562-b3fc-2c963f66afa7"



"""VC-API routes web requests schemas."""

from marshmallow import fields

from ....messaging.models.openapi import OpenAPISchema
from ..validation_result import PresentationVerificationResultSchema
from .credential import CredentialSchema, VerifiableCredentialSchema
from .options import LDProofVCOptionsSchema
from .presentation import PresentationSchema, VerifiablePresentationSchema


class ListCredentialsResponse(OpenAPISchema):
    """Response schema for listing credentials."""

    results = fields.List(fields.Nested(VerifiableCredentialSchema))


class FetchCredentialResponse(OpenAPISchema):
    """Response schema for fetching a credential."""

    results = fields.Nested(VerifiableCredentialSchema)


class IssueCredentialRequest(OpenAPISchema):
    """Request schema for issuing a credential."""

    credential = fields.Nested(CredentialSchema)
    options = fields.Nested(LDProofVCOptionsSchema)


class IssueCredentialResponse(OpenAPISchema):
    """Request schema for issuing a credential."""

    verifiableCredential = fields.Nested(VerifiableCredentialSchema)


class VerifyCredentialRequest(OpenAPISchema):
    """Request schema for verifying a credential."""

    verifiableCredential = fields.Nested(VerifiableCredentialSchema)
    options = fields.Nested(LDProofVCOptionsSchema)


class VerifyCredentialResponse(OpenAPISchema):
    """Request schema for verifying an LDP VP."""

    results = fields.Nested(PresentationVerificationResultSchema)


class StoreCredentialRequest(OpenAPISchema):
    """Request schema for verifying an LDP VP."""

    verifiableCredential = fields.Nested(VerifiableCredentialSchema)


class StoreCredentialResponse(OpenAPISchema):
    """Request schema for verifying an LDP VP."""

    credentialId = fields.Str()


class ProvePresentationRequest(OpenAPISchema):
    """Request schema for proving a presentation."""

    presentation = fields.Nested(PresentationSchema)
    options = fields.Nested(LDProofVCOptionsSchema)


class ProvePresentationResponse(OpenAPISchema):
    """Request schema for proving a presentation."""

    verifiablePresentation = fields.Nested(VerifiablePresentationSchema)


class VerifyPresentationRequest(OpenAPISchema):
    """Request schema for verifying a credential."""

    verifiablePresentation = fields.Nested(VerifiablePresentationSchema)
    options = fields.Nested(LDProofVCOptionsSchema)


class VerifyPresentationResponse(OpenAPISchema):
    """Request schema for verifying an LDP VP."""

    results = fields.Nested(PresentationVerificationResultSchema)



from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.base import BaseWallet
from .. import routes as test_module
from . import CRED_DEF_ID


class TestCredentialRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_credential_exchange_list(self):
        self.request.query = {
            "thread_id": "dummy",
            "connection_id": "dummy",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.query = mock.CoroutineMock()
            mock_cred_ex.query.return_value = [mock_cred_ex]
            mock_cred_ex.serialize = mock.MagicMock()
            mock_cred_ex.serialize.return_value = {"hello": "world"}

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.credential_exchange_list(self.request)
                mock_response.assert_called_once_with(
                    {"results": [mock_cred_ex.serialize.return_value]}
                )

    async def test_credential_exchange_list_x(self):
        self.request.query = {
            "thread_id": "dummy",
            "connection_id": "dummy",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.query = mock.CoroutineMock(
                side_effect=test_module.StorageError()
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_list(self.request)

    async def test_credential_exchange_retrieve(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value = mock_cred_ex
            mock_cred_ex.serialize = mock.MagicMock()
            mock_cred_ex.serialize.return_value = {"hello": "world"}

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.credential_exchange_retrieve(self.request)
                mock_response.assert_called_once_with(mock_cred_ex.serialize.return_value)

    async def test_credential_exchange_retrieve_not_found(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_retrieve(self.request)

    async def test_credential_exchange_retrieve_x(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value = mock_cred_ex
            mock_cred_ex.serialize = mock.MagicMock(
                side_effect=test_module.BaseModelError()
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_retrieve(self.request)

    async def test_credential_exchange_create(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module.CredentialPreview, "deserialize", autospec=True
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()

            mock_credential_manager.return_value.create_offer.return_value = (
                mock.CoroutineMock(),
                mock.CoroutineMock(),
            )

            mock_cred_ex_record = mock.MagicMock()
            mock_cred_offer = mock.MagicMock()

            mock_credential_manager.return_value.prepare_send.return_value = (
                mock_cred_ex_record,
                mock_cred_offer,
            )

            await test_module.credential_exchange_create(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_create_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module.CredentialPreview, "deserialize", autospec=True
            ),
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()

            mock_credential_manager.return_value.create_offer.return_value = (
                mock.CoroutineMock(),
                mock.CoroutineMock(),
            )

            mock_credential_manager.return_value.prepare_send.side_effect = (
                test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_create(self.request)

    async def test_credential_exchange_create_no_proposal(self):
        conn_id = "connection-id"

        self.request.json = mock.CoroutineMock(return_value={"connection_id": conn_id})

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_create(self.request)
        assert "credential_proposal" in str(context.exception)

    async def test_credential_exchange_send(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module.CredentialPreview, "deserialize", autospec=True
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()

            mock_credential_manager.return_value.create_offer.return_value = (
                mock.CoroutineMock(),
                mock.CoroutineMock(),
            )

            mock_cred_ex_record = mock.MagicMock()
            mock_cred_offer = mock.MagicMock()

            mock_credential_manager.return_value.prepare_send.return_value = (
                mock_cred_ex_record,
                mock_cred_offer,
            )

            await test_module.credential_exchange_send(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_send_no_proposal(self):
        conn_id = "connection-id"

        self.request.json = mock.CoroutineMock(return_value={"connection_id": conn_id})

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_send(self.request)
        assert "credential_proposal" in str(context.exception)

    async def test_credential_exchange_send_no_conn_record(self):
        conn_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": conn_id, "credential_proposal": preview_spec}
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send(self.request)

    async def test_credential_exchange_send_not_ready(self):
        conn_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": conn_id, "credential_proposal": preview_spec}
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send(self.request)

    async def test_credential_exchange_send_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module.CredentialPreview, "deserialize", autospec=True
            ),
        ):
            mock_cred_ex_record = mock.MagicMock(
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cred_offer = mock.MagicMock()

            mock_credential_manager.return_value = mock.MagicMock(
                create_offer=mock.CoroutineMock(
                    return_value=(
                        mock.CoroutineMock(),
                        mock.CoroutineMock(),
                    )
                ),
                prepare_send=mock.CoroutineMock(
                    return_value=(
                        mock_cred_ex_record,
                        mock_cred_offer,
                    )
                ),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send(self.request)

    async def test_credential_exchange_send_proposal(self):
        conn_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": conn_id, "credential_proposal": preview_spec}
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex_record = mock.MagicMock()
            mock_credential_manager.return_value.create_proposal.return_value = (
                mock_cred_ex_record
            )
            await test_module.credential_exchange_send_proposal(self.request)

            self.request["outbound_message_router"].assert_awaited_once_with(
                mock_cred_ex_record.credential_proposal_dict, connection_id=conn_id
            )
            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_send_proposal_no_conn_record(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module.CredentialPreview, "deserialize", autospec=True
            ),
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.create_proposal.return_value = (
                mock.MagicMock()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_send_proposal_deser_x(self):
        conn_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": conn_id, "credential_proposal": preview_spec}
        )

        with mock.patch.object(
            test_module.CredentialPreview, "deserialize", autospec=True
        ) as mock_preview_deser:
            mock_preview_deser.side_effect = test_module.BaseModelError()
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_send_proposal_not_ready(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module.CredentialPreview, "deserialize", autospec=True
            ),
        ):
            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_credential_manager.return_value.create_proposal.return_value = (
                mock.MagicMock()
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_send_proposal_x(self):
        conn_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": conn_id, "credential_proposal": preview_spec}
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            mock_cred_ex_record = mock.MagicMock(
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_credential_manager.return_value.create_proposal.return_value = (
                mock_cred_ex_record
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_create_free_offer(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        self.context.update_settings({"debug.auto_respond_credential_offer": True})

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_cred_ex_record = mock.MagicMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_create_free_offer(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_create_free_offer_no_cred_def_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credential_exchange_create_free_offer(self.request)

    async def test_credential_exchange_create_free_offer_no_preview(self):
        self.request.json = mock.CoroutineMock()
        self.request.json.return_value = {"comment": "comment", "cred_def_id": "dummy"}

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credential_exchange_create_free_offer(self.request)

    async def test_credential_exchange_create_free_offer_no_conn_id_no_public_did(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        self.context.update_settings({"default_endpoint": "http://1.2.3.4:8081"})
        self.session_inject[BaseWallet] = mock.MagicMock(
            get_public_did=mock.CoroutineMock(return_value=None),
        )

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credential_exchange_create_free_offer(self.request)

    async def test_credential_exchange_create_free_offer_deser_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.side_effect = (
                test_module.BaseModelError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_create_free_offer(self.request)

    async def test_credential_exchange_create_free_offer_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            mock_cred_ex_record = mock.MagicMock(
                serialize=mock.MagicMock(
                    side_effect=test_module.BaseModelError(),
                ),
                save_error_state=mock.CoroutineMock(),
            )
            mock_credential_manager.return_value = mock.MagicMock(
                create_offer=mock.CoroutineMock(
                    return_value=(
                        mock_cred_ex_record,
                        mock.MagicMock(),
                    )
                )
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_create_free_offer(self.request)

    async def test_credential_exchange_send_free_offer(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.create_offer.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_free_offer(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_send_free_offer_no_cred_def_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.json.return_value = {
            "comment": "comment",
            "credential_preview": "dummy",
        }

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_free_offer_no_preview(self):
        self.request.json = mock.CoroutineMock()
        self.request.json.return_value = {
            "comment": "comment",
            "cred_def_id": CRED_DEF_ID,
        }

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_free_offer_no_conn_record(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": "dummy",
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_free_offer_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.json.return_value["auto_issue"] = True

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
        ):
            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_free_offer_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "cred_def_id": CRED_DEF_ID,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_cred_ex_record = mock.MagicMock(
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )

            mock_credential_manager.return_value = mock.MagicMock(
                create_offer=mock.CoroutineMock(
                    return_value=(
                        mock_cred_ex_record,
                        mock.MagicMock(),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_bound_offer(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_PROPOSAL_RECEIVED
            )

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.create_offer.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_bound_offer(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_send_bound_offer_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_bound_offer_no_conn_record(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_cred_ex.STATE_PROPOSAL_RECEIVED,
                    save_error_state=mock.CoroutineMock(),
                )
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_bound_offer_bad_state(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_cred_ex.STATE_ACKED,
                    save_error_state=mock.CoroutineMock(),
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_bound_offer_not_ready(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_PROPOSAL_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_request(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_OFFER_RECEIVED
            )

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.create_request.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_request(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_send_request_no_conn(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "OobRecord", autospec=True) as mock_oob_rec,
            mock.patch.object(
                test_module, "default_did_from_verkey", autospec=True
            ) as mock_default_did_from_verkey,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_oob_rec.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock.MagicMock(our_recipient_key="our-recipient_key")
            )
            mock_default_did_from_verkey.return_value = "holder-did"

            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_OFFER_RECEIVED
            )
            mock_cred_ex.retrieve_by_id.return_value.connection_id = None

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.create_request.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_request(self.request)

            mock_credential_manager.return_value.create_request.assert_called_once_with(
                mock_cred_ex.retrieve_by_id.return_value, "holder-did"
            )
            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )
            mock_default_did_from_verkey.assert_called_once_with("our-recipient_key")

    async def test_credential_exchange_send_request_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_send_request(self.request)

    async def test_credential_exchange_send_request_no_conn_record(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value = mock.MagicMock(
                state=mock_cred_ex.STATE_OFFER_RECEIVED,
                save_error_state=mock.CoroutineMock(),
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_request(self.request)

    async def test_credential_exchange_send_request_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_OFFER_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_credential_manager.return_value.create_offer = mock.CoroutineMock()
            mock_credential_manager.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_request(self.request)

    async def test_credential_exchange_issue(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_REQUEST_RECEIVED
            )

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.issue_credential.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_issue(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_issue_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_no_conn_record(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cred_ex_rec = mock.MagicMock(
            connection_id="dummy",
            serialize=mock.MagicMock(),
            save_error_state=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex_cls,
        ):
            mock_cred_ex_rec.state = mock_cred_ex_cls.STATE_REQUEST_RECEIVED
            mock_cred_ex_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_cred_ex_rec
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.issue_credential = mock.CoroutineMock()
            mock_credential_manager.return_value.issue_credential.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_REQUEST_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_credential_manager.return_value.issue_credential = mock.CoroutineMock()
            mock_credential_manager.return_value.issue_credential.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_rev_reg_full(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cred_ex_rec = mock.MagicMock(
            connection_id="dummy",
            serialize=mock.MagicMock(),
            save_error_state=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex_cls,
        ):
            mock_cred_ex_cls.state = mock_cred_ex_cls.STATE_REQUEST_RECEIVED
            mock_cred_ex_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_cred_ex_rec
            )

            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = True

            mock_issue_cred = mock.CoroutineMock(
                side_effect=test_module.IndyIssuerError()
            )
            mock_credential_manager.return_value.issue_credential = mock_issue_cred

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_deser_x(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cred_ex_rec = mock.MagicMock(
            connection_id="dummy",
            serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
            save_error_state=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex_cls,
        ):
            mock_cred_ex_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_cred_ex_rec
            )
            mock_credential_manager.return_value = mock.MagicMock(
                issue_credential=mock.CoroutineMock(
                    return_value=(
                        mock_cred_ex_rec,
                        mock.MagicMock(),
                    )
                )
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_store(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_CREDENTIAL_RECEIVED
            )

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.store_credential.return_value = (
                mock_cred_ex_record
            )
            mock_credential_manager.return_value.send_credential_ack.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_store(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_store_bad_cred_id_json(self):
        self.request.json = mock.CoroutineMock(
            side_effect=test_module.JSONDecodeError("Nope", "Nope", 0)
        )
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_CREDENTIAL_RECEIVED
            )

            mock_cred_ex_record = mock.MagicMock()

            mock_credential_manager.return_value.store_credential.return_value = (
                mock_cred_ex_record
            )
            mock_credential_manager.return_value.send_credential_ack.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_store(self.request)

            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )

    async def test_credential_exchange_store_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_store_no_conn_record(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_cred_ex.STATE_CREDENTIAL_RECEIVED,
                    save_error_state=mock.CoroutineMock(),
                )
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_credential_manager.return_value.store_credential.return_value = (
                mock_cred_ex
            )
            mock_credential_manager.return_value.send_credential_ack.return_value = (
                mock_cred_ex,
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_store_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(test_module, "CredentialManager", autospec=True),
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.connection_id = "conn-123"
            mock_cred_ex.thread_id = "conn-123"
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_CREDENTIAL_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_store_x(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "CredentialManager", autospec=True
            ) as mock_credential_manager,
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex_cls,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_cred_ex_record = mock.MagicMock(
                state=mock_cred_ex_cls.STATE_CREDENTIAL_RECEIVED,
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cred_ex_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )

            mock_credential_manager.return_value = mock.MagicMock(
                store_credential=mock.CoroutineMock(return_value=mock_cred_ex_record),
                send_credential_ack=mock.CoroutineMock(
                    return_value=(mock_cred_ex_record, mock.MagicMock())
                ),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_remove(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value = mock_cred_ex

            mock_cred_ex.delete_record = mock.CoroutineMock()

            await test_module.credential_exchange_remove(self.request)

            mock_response.assert_called_once_with({})

    async def test_credential_exchange_remove_bad_cred_ex_id(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            # Emulate storage not found (bad cred ex id)
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_remove(self.request)

    async def test_credential_exchange_remove_x(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            # Emulate storage not found (bad cred ex id)
            mock_rec = mock.MagicMock(
                delete_record=mock.CoroutineMock(side_effect=test_module.StorageError())
            )
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(return_value=mock_rec)

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_remove(self.request)

    async def test_credential_exchange_problem_report(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"cred_ex_id": "dummy"}
        magic_report = mock.MagicMock()

        with (
            mock.patch.object(test_module, "CredentialManager", autospec=True),
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(
                test_module, "problem_report_for_record", mock.MagicMock()
            ) as mock_problem_report,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_problem_report.return_value = magic_report

            await test_module.credential_exchange_problem_report(self.request)

            self.request["outbound_message_router"].assert_awaited_once_with(
                magic_report,
                connection_id=mock_cred_ex.retrieve_by_id.return_value.connection_id,
            )
            mock_response.assert_called_once_with({})

    async def test_credential_exchange_problem_report_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V10CredentialExchange", autospec=True
        ) as mock_cred_ex:
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_problem_report(self.request)

    async def test_credential_exchange_problem_report_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "CredentialManager", autospec=True),
            mock.patch.object(test_module, "problem_report_for_record", mock.MagicMock()),
            mock.patch.object(
                test_module, "V10CredentialExchange", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    save_error_state=mock.CoroutineMock(
                        side_effect=test_module.StorageError()
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_problem_report(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Cred request artifacts to attach to RFC 453 messages."""

from typing import Optional, Sequence, Union

from marshmallow import EXCLUDE, fields

from .......indy.models.cred_abstract import IndyKeyCorrectnessProofSchema
from .......messaging.models.base import BaseModel, BaseModelSchema
from .......messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
)
from .......vc.vc_ld.models.credential import CredentialSchema, VerifiableCredential


class AnonCredsLinkSecret(BaseModel):
    """AnonCreds Link Secret Model."""

    class Meta:
        """AnonCredsLinkSecret metadata."""

        schema_class = "AnonCredsLinkSecretSchema"

    def __init__(
        self,
        nonce: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        key_correctness_proof: Optional[str] = None,
        **kwargs,
    ):
        """Initialize values for AnonCredsLinkSecret."""
        super().__init__(**kwargs)
        self.nonce = nonce
        self.cred_def_id = cred_def_id
        self.key_correctness_proof = key_correctness_proof


class AnonCredsLinkSecretSchema(BaseModelSchema):
    """AnonCreds Link Secret Schema."""

    class Meta:
        """AnonCredsLinkSecret schema metadata."""

        model_class = AnonCredsLinkSecret
        unknown = EXCLUDE

    nonce = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Nonce in credential abstract",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )

    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )

    key_correctness_proof = fields.Nested(
        IndyKeyCorrectnessProofSchema(),
        required=True,
        metadata={"description": "Key correctness proof"},
    )


class DidcommSignedAttachment(BaseModel):
    """Didcomm Signed Attachment Model."""

    class Meta:
        """DidcommSignedAttachment metadata."""

        schema_class = "DidcommSignedAttachmentSchema"

    def __init__(
        self,
        algs_supported: Sequence[str] = None,
        did_methods_supported: Sequence[str] = None,
        nonce: Optional[str] = None,
        **kwargs,
    ):
        """Initialize values for DidcommSignedAttachment."""
        super().__init__(**kwargs)
        self.algs_supported = algs_supported
        self.did_methods_supported = did_methods_supported
        self.nonce = nonce


class DidcommSignedAttachmentSchema(BaseModelSchema):
    """Didcomm Signed Attachment Schema."""

    class Meta:
        """Didcomm signed attachment schema metadata."""

        model_class = DidcommSignedAttachment
        unknown = EXCLUDE

    algs_supported = fields.List(fields.Str(), required=True)

    did_methods_supported = fields.List(fields.Str(), required=True)

    nonce = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Nonce in credential abstract",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )


class BindingMethod(BaseModel):
    """Binding Method Model."""

    class Meta:
        """Binding method metadata."""

        schema_class = "BindingMethodSchema"

    def __init__(
        self,
        anoncreds_link_secret: Union[dict, AnonCredsLinkSecret] = None,
        didcomm_signed_attachment: Union[dict, DidcommSignedAttachment] = None,
        **kwargs,
    ):
        """Initialize values for DidcommSignedAttachment."""
        super().__init__(**kwargs)
        self.anoncreds_link_secret = anoncreds_link_secret
        self.didcomm_signed_attachment = didcomm_signed_attachment


class BindingMethodSchema(BaseModelSchema):
    """VCDI Binding Method Schema."""

    class Meta:
        """VCDI binding method schema metadata."""

        model_class = BindingMethod
        unknown = EXCLUDE

    anoncreds_link_secret = fields.Nested(AnonCredsLinkSecretSchema, required=False)
    didcomm_signed_attachment = fields.Nested(
        DidcommSignedAttachmentSchema, required=True
    )


class VCDICredAbstract(BaseModel):
    """VCDI Credential Abstract."""

    class Meta:
        """VCDI credential abstract metadata."""

        schema_class = "VCDICredAbstractSchema"

    def __init__(
        self,
        data_model_versions_supported: Sequence[str] = None,
        binding_required: Optional[str] = None,
        binding_method: Optional[str] = None,
        credential: Union[dict, VerifiableCredential] = None,
        **kwargs,
    ):
        """Initialize vcdi cred abstract object.

        Args:
            data_model_versions_supported: supported versions for data model
            binding_required: boolean value
            binding_method: required if binding_required is true
            credential: credential object
            kwargs: additional key-value arguments to map into message class properties
        """
        super().__init__(**kwargs)
        self.data_model_versions_supported = data_model_versions_supported
        self.binding_required = binding_required
        self.binding_method = binding_method
        self.credential = credential


class VCDICredAbstractSchema(BaseModelSchema):
    """VCDI Credential Abstract Schema."""

    class Meta:
        """VCDICredAbstractSchema metadata."""

        model_class = VCDICredAbstract
        unknown = EXCLUDE

    data_model_versions_supported = fields.List(
        fields.Str(), required=True, metadata={"description": "", "example": ""}
    )

    binding_required = fields.Bool(
        required=False, metadata={"description": "", "example": ""}
    )

    binding_method = fields.Nested(
        BindingMethodSchema(),
        required=binding_required,
        metadata={"description": "", "example": ""},
    )

    credential = fields.Nested(
        CredentialSchema(),
        required=True,
        metadata={"description": "", "example": ""},
    )
